/*
 *  AUTHOR: Sven Goldt (goldt@math.tu-berlin.de)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 */

/*
 Egy kozismert, es egyszeru pelda program a soros terminal keszitesere.
*/

#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <stdlib.h>
#include <sys/wait.h>

#define BAUDRATE B38400
#define MODEMDEVICE "/dev/modem"
#define ENDMINITERM 2 /* ctrl-b kilepes a programbol */

#define _POSIX_SOURCE 1 /* POSIX kompatibilis forras */

#define FALSE 0
#define TRUE 1

volatile int STOP=FALSE; 

void child_handler(int s)
{
  STOP=TRUE;
}

int main(void)
{
  int fd,c;
  struct termios oldtio,newtio,oldstdtio,newstdtio;
  struct sigaction sa;
 
  /* Megnyitjuk a modem eszkozt olvasasra es irasra. Azonban nem ugy, hogy
     ne ertelmezze a terminal kontrol karaktereket. */
  fd = open(MODEMDEVICE, O_RDWR | O_NOCTTY);
  if (fd <0) {perror(MODEMDEVICE); exit(-1); }

  /* Elmentjuk az aktualis modem beallitasokat. */ 
  tcgetattr(fd,&oldtio);
 
  /* 
  BAUDRATE: a soros port sebessege
  CRTSCTS : a port hardver atvitel szabalyozasa
  CS8     : 8n1 (8 bit, nincs paritas,1 stopbit)
  CLOCAL  : lokalis kapcsolat, nincsenek modem jelek
  CREAD   : a karakterek fogadasanak bekapcsolasa
  */
  newtio.c_cflag = BAUDRATE | CRTSCTS | CS8 | CLOCAL | CREAD;
 
  /*
  IGNPAR  : a paritas hibas bajtokat figyelmen kivul hagyja
  */
  newtio.c_iflag = IGNPAR;
 
  /* Nyers kimenet beallitasa. */
  newtio.c_oflag = 0;
 
  /* Az input beallitas nem kanonikus es a visszejelzest is 
     kikapcsoljuk. */        
  newtio.c_lflag = 0;
 
  /* Az olvasas blokkolodik, amig 1 karakter erkezik. */
  newtio.c_cc[VMIN]=1;
  newtio.c_cc[VTIME]=0;

  /* Toroljuk soros port buffereit es aktivaljuk a beallitasokat. */
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd,TCSANOW,&newtio);

  /* Az stdout beallitasokat a soros port beallitasokkal megegyezore
     allitjuk be. (Ez a sor egy regebbi linux hiba miatt kerult be. */
  tcsetattr(1,TCSANOW,&newtio);
 
  /* Kikapcsolja az stdin bufferezeset es visszajelzeset. */
  tcgetattr(0,&oldstdtio);
  tcgetattr(0,&newstdtio);
  newstdtio.c_lflag &= ~(ICANON | ECHO);
  tcsetattr(0,TCSANOW,&newstdtio);

  /* A terminal beallitasok megtortentek. A kovetkezo a
     bemenet/kimenet kezelese. */
  switch (fork())
  {
    case 0: /* gyerek */
      /* A felhasznaloi bemenet */

      /* A kimenetet lezarjuk. */
      close(1); 

      for (c=getchar(); c!= ENDMINITERM ; c=getchar()) write(fd,&c,1);

      /* Visszaallitjuk a modem beallitasokat. */
      tcsetattr(fd,TCSANOW,&oldtio); 
      /* Visszaallitjuk a terminal beallitasokat. */
      tcsetattr(0,TCSANOW,&oldstdtio);
      close(fd);
      exit(0); /* egy SIGCHLD szignalt kuld a szulonek */
      break;

    case -1:
      /* A fork() hiba lekezelese. */
      perror("fork");
      /*Visszaallitjuk a modem beallitasokat. */
      tcsetattr(fd,TCSANOW,&oldtio);
      close(fd);
      exit(-1);

    default: /* szulo */
      /* A modem olvasasa. */

      /* A bemenetet lezarjuk. */
      close(0); 

      /* A SIGCHLD szignal lekezelese. */
      sa.sa_handler = child_handler;
      sa.sa_flags = 0;
      sigaction(SIGCHLD,&sa,NULL);
      /* A kilepesig olvasunk. */
      while (STOP==FALSE)
      {
	/* Olvasas a modemrol. */
        read(fd,&c,1);
        /* Kiiras az stdout-ra. */
        write(1,&c,1);
      }
      
      /* Minden gyerek processzt megvarunk, hogy ne valjanak
         zombie-va. */
      wait(NULL);
      break;
  }

  return 0;
}
